/*
 * puzzle.c
 *
 *
 * reconstruction of tree topologies with quartet puzzling
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 *
 */


#define MAIN_MODULE 1

#include "puzzle.h"


/******************* internal functions ****************************************/

/* set options interactively */
void setoptions()
{	
	char inputline[10];	


	/* defaults */
	Info_optn   = FALSE; /* show status of computation     */
	first_optn  = FALSE; /* special care for first quartet */
	Approx_optn = TRUE;  /* approximate likelihoods        */
	Numtrial = 1000;     /* number of Puzzle trees         */


#ifdef NUC	
	AlphaYR = 1.0;
	AlphaBeta = 4.0;
	Frequ_optn = TRUE;
#endif

#ifdef PROT
	Dayhf_optn = FALSE;
	Jtt_optn = TRUE;
	Poisn_optn = FALSE;
	Frequ_optn = FALSE;
#endif	

	fprintf(stderr, "\nReconstructing tree topologies with quartet puzzling\n");

 	for (;;){
			
		fprintf(stderr, "\n\nSettings for this run of Puzzle version %s (%s):\n\n",
			VERSION, MODEL);
#ifdef NUC			
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if(AlphaBeta == 1.0 && AlphaYR == 1.0) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
		if((AlphaBeta != 1.0 && AlphaYR == 1.0)) fprintf(stderr, "%s\n",
			(Frequ_optn ? "HKY (Hasegawa et al. 1985)" : "K2P (Kimura 1980)"));	
		if(AlphaYR != 1.0) fprintf(stderr, "TN (Tamura-Nei 1993)\n");
		fprintf(stderr, "t    Transition rate/transversion rate?  %.2f\n",
			AlphaBeta);	
		if (AlphaYR != 1.0)
			fprintf(stderr, "p  Transition Y rate/transition R rate?  %.2f\n",
				AlphaYR);
#endif

#ifdef PROT
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if (Dayhf_optn) fprintf(stderr, "Dayhoff (Dayhoff et al. 1978)\n");	
		if (Jtt_optn) fprintf(stderr, "JTT (Jones et al. 1992)\n");
		if (Poisn_optn) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
#endif

		fprintf(stderr, "f                 Use data frequencies?  %s\n",
			(Frequ_optn ? "Yes" : "No") );	
		fprintf(stderr, "a       Approximate maximum likelihood?  %s\n",
				(Approx_optn ? "Yes" : "No"));
		fprintf(stderr, "n               Number of Puzzle trees?  %d\n",
				Numtrial);
		fprintf(stderr, "s       Special care for first quartet?  %s\n",
				(first_optn ? "Yes" : "No"));
		fprintf(stderr, "i          Show details of computation?  %s\n",
				(Info_optn ? "Yes" : "No"));
		
		fprintf(stderr, "\nStart with analysis? ");
		fprintf(stderr, "(type y or choose from the menu) ");
		
		gets(inputline);

		if(tolower(inputline[0]) == 'y')
		break;

		switch (tolower(inputline[0])) {

			case 'a':	Approx_optn = 1 - Approx_optn;
						break;
			
			case 'i':	Info_optn = 1 - Info_optn;
						break;
			
			case 's':	first_optn = 1 - first_optn;
						break;
						
			case 'n':	fprintf(stderr, "\n\nNumber of Puzzle trees: ");
						scanf("%d", &Numtrial);
						gets(inputline);
						if (Numtrial < 1) {
							fprintf(stderr, "\n\nThis is no valid input!\n");
							Numtrial = 1000;
						}
						break;
#ifdef NUC			
			case 'm':	if(AlphaBeta == 1.0 && AlphaYR == 1.0) {/* -> K2P/HKY */
							AlphaBeta = 4.0;
							AlphaYR = 1.0;
							break;
						}
						if(AlphaBeta != 1.0 && AlphaYR == 1.0) {/* -> TN */
							AlphaBeta = 4.0;
							AlphaYR = 2.0;
							break;
						}
						if(AlphaYR != 1.0) { /* -> JC */
							AlphaBeta = 1.0;
							AlphaYR = 1.0;
							break;
						}
						break;
						
	
			case 't':	fprintf(stderr, "\n\nEnter new ratio for transition"); 
						fprintf(stderr, " rate/transversion rate: ");
						scanf("%lf", &AlphaBeta);
						gets(inputline);
						break;
						
				
			case 'p':	if (AlphaYR == 1.0) {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						} else {
							fprintf(stderr, "\n\nEnter new ratio for transition");
							fprintf(stderr, " Y rate/transition R rate: ");
							scanf("%lf", &AlphaYR);
							gets(inputline);
							break;
						}
#endif

#ifdef PROT
			case 'm':	if (Dayhf_optn) { /* -> JC/F */
							Dayhf_optn = FALSE;
							Jtt_optn = FALSE;
							Poisn_optn = TRUE;
							break;
						}
						if (Poisn_optn) { /* -> JTT model */
							Dayhf_optn = FALSE;
							Jtt_optn = TRUE;
							Poisn_optn = FALSE;
							break;
						}
						if (Jtt_optn) { /* -> Dayhoff model */
							Dayhf_optn = TRUE;
							Jtt_optn = FALSE;
							Poisn_optn = FALSE;
							break;
						}
						break;

#endif			
		case 'f':	Frequ_optn = 1 - Frequ_optn;
						break;
					
		default:	fprintf(stderr, "\n\nThis is not a possible option!\n");
		}
	}
	fprintf(stderr, "\n\n");
}


/* open all needed files */
void opentreefile()
{	
	char inputline[64];

	fprintf(stderr, "\nPlease enter the name for the output file: ");
	gets(inputline);
	if ((treefp = fopen(inputline, "w")) == NULL) {
		fprintf(stderr, "\nERROR: I can't create the output file: %s\n", inputline);
		fprintf(stderr, "Please enter another name for this file: ");
		gets(inputline);
		fprintf(stderr, "\n\n");
		if ((treefp = fopen(inputline, "w")) == NULL) {
			fprintf(stderr, "\nERROR: I could not create this file either!\n");
			exit(1);
		}
	}
}


/* close needed files */
void closetreefile()
{	
	fclose(treefp);
}



/******************************** main module *********************************/


main()
{	
	int i, j, a, b, c, k, ID;
	unsigned long nq;
	time_t starttime, stoptime;


	/* set options in interaction with the user */
	setoptions();

	/* initialize random numbers generator */
	initrandom();

	/* open treefile */
	opentreefile();

	/* start timer */
	time(&starttime);
	
	/* initialise ML */
	fprintf(stderr, "\n\nINFO: Processing sequence input data\n");
	qmlstart();
	
	/* prepare for randomizing input order */
	trueID = new_ivector(Maxspc);
	freeID = new_ivector(Maxspc);
			
	/* number of branches */
	Maxbrnch = 2 * Maxspc - 3; 

	/* allocate memory for quartets */
	mallocquartets(Maxspc);
			
	/* compute quartets */
	fprintf(stderr, "\nINFO: Computing quartet likelihoods\n");
	computelihoods();
	
	fprintf(stderr, "\nINFO: Starting quartet puzzling analysis\n");
	
	for (Currtrial = 0; Currtrial < Numtrial; Currtrial++) {
				
		/* randomize input order */
				
		for (k = 0; k < Maxspc; k++) freeID[k] = TRUE;
			if (first_optn) {
				trueID[0] = startA;
				freeID[startA] = FALSE;
				trueID[1] = startB;
				freeID[startB] = FALSE;
				trueID[2] = startC;
				freeID[startC] = FALSE;
				trueID[3] = startD;
				freeID[startD] = FALSE;
				for (k = 4; k < Maxspc; k++) {
					/* random integer in the range [0, Maxspc-1-k] */
					ID = randominteger(Maxspc - k);
					i = -1;
					j = -1;
					do {
						i++;
						if (freeID[i] == TRUE) j++; 
					} while ( j != ID);
						trueID[k] = i;
						freeID[i] = FALSE;												
				}
			} else {
				for (k = 0; k < Maxspc; k++) {
					/* random integer in the range [0, Maxspc-1-k] */
					ID = randominteger(Maxspc - k);
						i = -1;
						j = -1;
						do {
							i++;
							if (freeID[i] == TRUE) j++; 
						} while ( j != ID);
							trueID[k] = i;
							freeID[i] = FALSE;												
				}
			}
					

			/* initialize tree */
			inittree();
			if (Info_optn) {
				fprintf(stderr, "\n  Quartet puzzling tree # %d:\n",
				Currtrial + 1);
				fprintf(stderr, "   leaf 0  > ");
				fputid10(stderr, trueID[0]);
				fprintf(stderr, " <  added  (0 quartets analysed)\n");
				fprintf(stderr, "   leaf 1  > ");
				fputid10(stderr, trueID[1]);
				fprintf(stderr, " <  added  (0 quartets analysed)\n");
				fprintf(stderr, "   leaf 2  > ");
				fputid10(stderr, trueID[2]);
				fprintf(stderr, " <  added  (0 quartets analysed)\n");
			}

			/* adding all other leafs */
			for (i = 3; i < Maxspc; i++) {
			
				/* determine place for next leaf */ 
						
				/* clear all edgeinfos */
				resetedgeinfo();
				/* clear counter of quartets */
				nq = 0;
		
					/* try all quartets */
					for (a = 0; a < nextleaf - 2; a++)
						for (b = a + 1; b < nextleaf - 1; b++)
							for (c = b + 1; c < nextleaf; c++) {
						
								/*	check which two leafs out of a, b, c
									are closer related to each other than
									to leaf i
									   
									These two are chooseA and chooseB */

									checkquartet(a, b, c, i);

									incrementedgeinfo(chooseA, chooseB, 1);
									nq++;
							}

				/* find out which edge has the lowest edgeinfo */
				minimumedgeinfo();	
				
				/* add the next leaf on this edge */
				addnextleaf(minedge);
			
				if(Info_optn) {
					fprintf(stderr, "   leaf %d  > ", i);
					fputid10(stderr, trueID[i]);
					fprintf(stderr, " <  added  (%lu quartets analysed)\n", nq);
					}
				}

				/* writing treefile */
				writetree(treefp);

				/* free memory before the next trial */
				freetree();
			}
			
	/* free memory */
	freequartets();
	free_ivector(trueID);
	free_ivector(freeID);
	
	/* close treefiles */
	fprintf(stderr, "\nINFO: Writing all %d trees to output file\n", Numtrial);
	closetreefile();
	
	/* finish ML */
	qmlfinish();
	
	/* stop timer */
	time(&stoptime);
	fprintf(stderr, "\nINFO: This run took %.0f seconds (= %.1f minutes)\n",
		(double) (stoptime - starttime), (stoptime - starttime)/60.);

	return;
}
